/* Dia -- an diagram creation/manipulation program
 * Copyright (C) 1998 Alexander Larsson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>

#include "intl.h"
#include "object.h"
#include "element.h"
#include "connectionpoint.h"
#include "render.h"
#include "attributes.h"
#include "widgets.h"
#include "properties.h"

#include "pixmaps/end.xpm"

#define END_DEFAULT_WIDTH 2.0
#define END_DEFAULT_HEIGHT 1.0
#define END_TEXT_BORDER_WIDTH_X 0.2
#define END_TEXT_BORDER_WIDTH_Y 0.3
#define END_FONT_HEIGHT 0.8

typedef struct _End End;

const char *name = "End";

struct _End {
  Element element;

  ConnectionPoint connections[8];

  real border_width;
  Color border_color;
  Color inner_color;
  
  DiaFont *font;
  real name_width;
  
};

static real end_distance_from(End *end, Point *point);
static void end_select(End *end, Point *clicked_point,
		       Renderer *interactive_renderer);
static void end_move_handle(End *end, Handle *handle,
			    Point *to, HandleMoveReason reason, ModifierKeys modifiers);
static void end_move(End *end, Point *to);
static void end_draw(End *end, Renderer *renderer);
static void end_update_data(End *end);
static Object *end_create(Point *startpoint,
			     void *user_data,
			     Handle **handle1,
			     Handle **handle2);
static void end_destroy(End *end);
static Object *end_copy(End *end);
static PropDescription *
end_describe_props(End *end);
static void
end_get_props(End *end, GPtrArray *props);
static void
end_set_props(End *end, GPtrArray *props);

static void end_save(End *end, ObjectNode obj_node,
			const char *filename);
static Object *end_load(ObjectNode obj_node, int version,
			   const char *filename);

static ObjectTypeOps end_type_ops =
{
  (CreateFunc) end_create,
  (LoadFunc)   end_load,
  (SaveFunc)   end_save
};

ObjectType end_type =
{
  "Cast - End",  /* name */
  0,                 /* version */
  (char **) end_xpm, /* pixmap */

  &end_type_ops      /* ops */
};

ObjectType *_end_type = (ObjectType *) &end_type;

static ObjectOps end_ops = {
  (DestroyFunc)         end_destroy,
  (DrawFunc)            end_draw,
  (DistanceFunc)        end_distance_from,
  (SelectFunc)          end_select,
  (CopyFunc)            end_copy,
  (MoveFunc)            end_move,
  (MoveHandleFunc)      end_move_handle,
  (GetPropertiesFunc)   object_create_props_dialog,
  (ApplyPropertiesFunc) object_apply_props_from_dialog,
  (ObjectMenuFunc)      NULL,
  (DescribePropsFunc)   end_describe_props,
  (GetPropsFunc)        end_get_props,
  (SetPropsFunc)        end_set_props
};

static PropDescription end_props[] = {
  ELEMENT_COMMON_PROPERTIES,
//  { "name", PROP_TYPE_STRING, PROP_FLAG_VISIBLE,
//    N_("Name:"), NULL, NULL },
//  { "weak", PROP_TYPE_BOOL, PROP_FLAG_VISIBLE,
//    N_("Weak:"), NULL, NULL },
  PROP_STD_LINE_WIDTH,
  PROP_STD_LINE_COLOUR,
  PROP_STD_FILL_COLOUR,
  PROP_DESC_END
};

static PropDescription *
end_describe_props(End *end)
{
  if (end_props[0].quark == 0)
    prop_desc_list_calculate_quarks(end_props);
  return end_props;
}

static PropOffset end_offsets[] = {
  ELEMENT_COMMON_PROPERTIES_OFFSETS,
//  { "name", PROP_TYPE_STRING, offsetof(End, name) },
//  { "weak", PROP_TYPE_BOOL, offsetof(End, weak) },
  { "line_width", PROP_TYPE_REAL, offsetof(End, border_width) },
  { "line_colour", PROP_TYPE_COLOUR, offsetof(End, border_color) },
  { "fill_colour", PROP_TYPE_COLOUR, offsetof(End, inner_color) },
  { NULL, 0, 0}
};


static void
end_get_props(End *end, GPtrArray *props)
{
  object_get_props_from_offsets(&end->element.object, 
                                end_offsets, props);
}

static void
end_set_props(End *end, GPtrArray *props)
{
  object_set_props_from_offsets(&end->element.object, 
                                end_offsets, props);
  end_update_data(end);
}

static real
end_distance_from(End *end, Point *point)
{
  Element *elem = &end->element;
  Rectangle rect;

  rect.left = elem->corner.x - end->border_width/2;
  rect.right = elem->corner.x + elem->width + end->border_width/2;
  rect.top = elem->corner.y - end->border_width/2;
  rect.bottom = elem->corner.y + elem->height + end->border_width/2;
  return distance_rectangle_point(&rect, point);
}

static void
end_select(End *end, Point *clicked_point,
	   Renderer *interactive_renderer)
{
  element_update_handles(&end->element);
}

static void
end_move_handle(End *end, Handle *handle,
		Point *to, HandleMoveReason reason, ModifierKeys modifiers)
{
  assert(end!=NULL);
  assert(handle!=NULL);
  assert(to!=NULL);

  element_move_handle(&end->element, handle->id, to, reason);

  end_update_data(end);
}

static void
end_move(End *end, Point *to)
{
  end->element.corner = *to;
  
  end_update_data(end);
}

static void end_draw(End *end, Renderer *renderer)
{
  Point ul_corner, lr_corner;
  Point p;
  Element *elem;
//  coord diff;

  assert(end != NULL);
  assert(renderer != NULL);

  elem = &end->element;

  ul_corner.x = elem->corner.x;
  ul_corner.y = elem->corner.y;
  lr_corner.x = elem->corner.x + elem->width;
  lr_corner.y = elem->corner.y + elem->height;
  
  renderer->ops->set_fillstyle(renderer, FILLSTYLE_SOLID);
  
  renderer->ops->fill_rect(renderer, 
			   &ul_corner,
			   &lr_corner, 
			   &end->inner_color);

  renderer->ops->set_linewidth(renderer, end->border_width);
  renderer->ops->set_linestyle(renderer, LINESTYLE_SOLID);
  renderer->ops->set_linejoin(renderer, LINEJOIN_MITER);

  renderer->ops->draw_rect(renderer, 
			   &ul_corner,
			   &lr_corner, 
			   &end->border_color);

/*  if(end->weak) {
    diff = WEAK_BORDER_WIDTH;
  //MIN(elem->width / 2.0 * 0.20, elem->height / 2.0 * 0.20)
    ul_corner.x += diff; 
    ul_corner.y += diff;
    lr_corner.x -= diff;
    lr_corner.y -= diff;
    renderer->ops->draw_rect(renderer, 
			     &ul_corner, &lr_corner,
			     &end->border_color);
  }
*/
  p.x = elem->corner.x + elem->width / 2.0;
  p.y = elem->corner.y + (elem->height - END_FONT_HEIGHT)/2.0 + font_ascent(end->font, END_FONT_HEIGHT);
  renderer->ops->set_font(renderer, 
			  end->font, END_FONT_HEIGHT);
  renderer->ops->draw_string(renderer, 
			     name, 
			     &p, ALIGN_CENTER, 
			     &color_black);
}

static void end_update_data(End *end)
{
  Element *elem = &end->element;
  Object *obj = &elem->object;
  ElementBBExtras *extra = &elem->extra_spacing;

  end->name_width =
    font_string_width(name, end->font, END_FONT_HEIGHT);

  elem->width = end->name_width + 2 * END_TEXT_BORDER_WIDTH_X;
  elem->height = END_FONT_HEIGHT + 2 * END_TEXT_BORDER_WIDTH_Y;

  /* Update connections: */
  end->connections[0].pos = elem->corner;
  end->connections[1].pos.x = elem->corner.x + elem->width / 2.0;
  end->connections[1].pos.y = elem->corner.y;
  end->connections[2].pos.x = elem->corner.x + elem->width;
  end->connections[2].pos.y = elem->corner.y;
  end->connections[3].pos.x = elem->corner.x;
  end->connections[3].pos.y = elem->corner.y + elem->height / 2.0;
  end->connections[4].pos.x = elem->corner.x + elem->width;
  end->connections[4].pos.y = elem->corner.y + elem->height / 2.0;
  end->connections[5].pos.x = elem->corner.x;
  end->connections[5].pos.y = elem->corner.y + elem->height;
  end->connections[6].pos.x = elem->corner.x + elem->width / 2.0;
  end->connections[6].pos.y = elem->corner.y + elem->height;
  end->connections[7].pos.x = elem->corner.x + elem->width;
  end->connections[7].pos.y = elem->corner.y + elem->height;

  extra->border_trans = end->border_width/2.0;
  element_update_boundingbox(elem);
  
  obj->position = elem->corner;
  
  element_update_handles(elem);
}

static Object *
end_create(Point *startpoint,
	      void *user_data,
	      Handle **handle1,
	      Handle **handle2)
{
  End *end;
  Element *elem;
  Object *obj;
  int i;

  end = g_malloc0(sizeof(End));
  elem = &end->element;
  obj = &elem->object;
  
  obj->type = &end_type;

  obj->ops = &end_ops;

  elem->corner = *startpoint;
  elem->width = END_DEFAULT_WIDTH;
  elem->height = END_DEFAULT_WIDTH;

  end->border_width =  attributes_get_default_linewidth();
  end->border_color = attributes_get_foreground();
  end->inner_color = attributes_get_background();
  
  element_init(elem, 8, 8);

  for (i=0;i<8;i++) {
    obj->connections[i] = &end->connections[i];
    end->connections[i].object = obj;
    end->connections[i].connected = NULL;
  }

//  end->weak = GPOINTER_TO_INT(user_data);
  end->font = font_getfont("Courier");
//  end->name = g_strdup(_("End"));

  end->name_width =
    font_string_width(name, end->font, END_FONT_HEIGHT);
  
  end_update_data(end);

  for (i=0;i<8;i++) {
    obj->handles[i]->type = HANDLE_NON_MOVABLE;
  }

  *handle1 = NULL;
  *handle2 = obj->handles[0];  
  return &end->element.object;
}

static void
end_destroy(End *end)
{
  element_destroy(&end->element);
//  g_free(end->name);
}

static Object *
end_copy(End *end)
{
  int i;
  End *newend;
  Element *elem, *newelem;
  Object *newobj;
  
  elem = &end->element;
  
  newend = g_malloc0(sizeof(End));
  newelem = &newend->element;
  newobj = &newelem->object;

  element_copy(elem, newelem);

  newend->border_width = end->border_width;
  newend->border_color = end->border_color;
  newend->inner_color = end->inner_color;
  
  for (i=0;i<8;i++) {
    newobj->connections[i] = &newend->connections[i];
    newend->connections[i].object = newobj;
    newend->connections[i].connected = NULL;
    newend->connections[i].pos = end->connections[i].pos;
    newend->connections[i].last_pos = end->connections[i].last_pos;
  }

  newend->font = end->font;
//  newend->name = g_strdup(end->name);
  newend->name_width = end->name_width;

//  newend->weak = end->weak;

  return &newend->element.object;
}

static void
end_save(End *end, ObjectNode obj_node, const char *filename)
{
  element_save(&end->element, obj_node);

  data_add_real(new_attribute(obj_node, "border_width"),
		end->border_width);
  data_add_color(new_attribute(obj_node, "border_color"),
		 &end->border_color);
  data_add_color(new_attribute(obj_node, "inner_color"),
		 &end->inner_color);
//  data_add_string(new_attribute(obj_node, "name"),
//		  end->name);
//  data_add_boolean(new_attribute(obj_node, "weak"),
//		   end->weak);
}

static Object *
end_load(ObjectNode obj_node, int version, const char *filename)
{
  End *end;
  Element *elem;
  Object *obj;
  int i;
  AttributeNode attr;

  end = g_malloc0(sizeof(End));
  elem = &end->element;
  obj = &elem->object;
  
  obj->type = &end_type;
  obj->ops = &end_ops;

  element_load(elem, obj_node);
  
  end->border_width = 0.1;
  attr = object_find_attribute(obj_node, "border_width");
  if (attr != NULL)
    end->border_width =  data_real( attribute_first_data(attr) );

  end->border_color = color_black;
  attr = object_find_attribute(obj_node, "border_color");
  if (attr != NULL)
    data_color(attribute_first_data(attr), &end->border_color);
  
  end->inner_color = color_white;
  attr = object_find_attribute(obj_node, "inner_color");
  if (attr != NULL)
    data_color(attribute_first_data(attr), &end->inner_color);
  
//  end->name = NULL;
//  attr = object_find_attribute(obj_node, "name");
//  if (attr != NULL)
//    end->name = data_string(attribute_first_data(attr));

//  attr = object_find_attribute(obj_node, "weak");
//  if (attr != NULL)
//    end->weak = data_boolean(attribute_first_data(attr));

  element_init(elem, 8, 8);

  for (i=0;i<8;i++) {
    obj->connections[i] = &end->connections[i];
    end->connections[i].object = obj;
    end->connections[i].connected = NULL;
  }

  end->font = font_getfont("Courier");

  end->name_width =
    font_string_width(name, end->font, END_FONT_HEIGHT);

  end_update_data(end);

  for (i=0;i<8;i++) {
    obj->handles[i]->type = HANDLE_NON_MOVABLE;
  }

  return &end->element.object;
}
