/* Dia -- an diagram creation/manipulation program
 * Copyright (C) 1998 Alexander Larsson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>

#include "intl.h"
#include "object.h"
#include "element.h"
#include "connectionpoint.h"
#include "render.h"
#include "attributes.h"
#include "widgets.h"
#include "properties.h"

#include "pixmaps/mul.xpm"

#include "cast.h"
#include "task.h"
#include "mul.h"
#include "mul_callbacks.h"



ObjectType mul_type;
ObjectType *_mul_type = (ObjectType *) &mul_type;


//mul functions
//mul_ops
static void mul_destroy(Mul *mul);
static void mul_draw(Mul *mul, Renderer *renderer);
static real mul_distance_from(Mul *mul, Point *point);
static void mul_select(Mul *mul, Point *clicked_point, Renderer *interactive_renderer);
static Object *mul_copy(Mul *mul);
static void mul_move(Mul *mul, Point *to);
static void mul_move_handle(Mul *mul, Handle *handle, Point *to, HandleMoveReason reason, ModifierKeys modifiers);

//mul_type_ops
static Object *mul_create(Point *startpoint, void *user_data, Handle **handle1, Handle **handle2);
static void mul_save(Mul *mul, ObjectNode obj_node, const char *filename);
static Object *mul_load(ObjectNode obj_node, int version, const char *filename);


//register mul functions
//mul_ops
static ObjectOps mul_ops = {
	(DestroyFunc)		mul_destroy,
	(DrawFunc)		mul_draw,
	(DistanceFunc)		mul_distance_from,
	(SelectFunc)		mul_select,
	(CopyFunc)		mul_copy,
	(MoveFunc)		mul_move,
	(MoveHandleFunc)	mul_move_handle,
	(GetPropertiesFunc)	mul_get_props,
	(ApplyPropertiesFunc)	mul_apply_props,
	(ObjectMenuFunc)	NULL
};


//mul_type_ops
static ObjectTypeOps mul_type_ops =
{
	(CreateFunc)	mul_create,
	(LoadFunc)	mul_load,
	(SaveFunc)	mul_save
};


//register mul (cf cast.c)
ObjectType mul_type =
{
	"Cast - Multiply",		/* name */
	0,			/* version */
	(char **) mul_xpm,	/* pixmap */

	&mul_type_ops		/* ops */
};

//register mul functions
//mul_ops
//removes a mul from memory
static void mul_destroy(Mul *mul)
{
	GSList *tasks;

	element_destroy(&mul->element);

//	task_destroy(mul->task[0]);
//	task_destroy(mul->task[1]);

	tasks = mul->tasks;
	do
	{
		task_destroy(tasks->data);
		tasks = g_slist_next(tasks);
	}
	while(tasks);

	if(mul->properties_dialog != NULL)
	{
		gtk_widget_destroy(mul->properties_dialog->dialog);
		g_free(mul->properties_dialog);
	}
}

//called to draw a mul
static void mul_draw(Mul *mul, Renderer *renderer)
{
	Task *task;
	GSList *tasks;
	Point ul_corner, lr_corner;
	Point p;
	Point p2;
	Element *elem;
	int i;

	assert(mul != NULL);
	assert(renderer != NULL);

	elem = &mul->element;

	//rectangles
	//outer
	ul_corner.x = elem->corner.x;
	ul_corner.y = elem->corner.y;
	lr_corner.x = elem->corner.x + elem->width;
	lr_corner.y = elem->corner.y + elem->height;
	renderer->ops->set_fillstyle(renderer, FILLSTYLE_SOLID);
	renderer->ops->fill_rect(renderer,
			&ul_corner,
			&lr_corner,
			&mul->inner_color);
	renderer->ops->set_linewidth(renderer, border_width);
	renderer->ops->set_linestyle(renderer, LINESTYLE_SOLID);
	renderer->ops->set_linejoin(renderer, LINEJOIN_MITER);
	renderer->ops->draw_rect(renderer,
			&ul_corner,
			&lr_corner,
			&mul->border_color);

	ul_corner.x = ul_corner.x + MUL_TEXT_BORDER_WIDTH_X;
	lr_corner.x = lr_corner.x - MUL_TEXT_BORDER_WIDTH_X;

	//tasks
	tasks = mul->tasks;
	i = 0;
	do
	{
		//FIXME drawing can be optimized A LOT (without i and multiplications)
		//line
		p.x = elem->corner.x;
		p.y = elem->corner.y + (i * (MUL_FONT_HEIGHT + MUL_TEXT_BORDER_WIDTH_Y + MUL_TEXT_BORDER_WIDTH_Y));
		p2.x = elem->corner.x + elem->width;
		p2.y = elem->corner.y + (i * (MUL_FONT_HEIGHT + MUL_TEXT_BORDER_WIDTH_Y + MUL_TEXT_BORDER_WIDTH_Y));
		renderer->ops->draw_line(renderer, &p, &p2, &mul->border_color);

		//task
		task = tasks->data;

		ul_corner.y = elem->corner.y + (i * (MUL_FONT_HEIGHT + MUL_TEXT_BORDER_WIDTH_Y + MUL_TEXT_BORDER_WIDTH_Y) + MUL_TEXT_BORDER_WIDTH_Y);
		lr_corner.y = elem->corner.y + ((i + 1) * (MUL_FONT_HEIGHT + MUL_TEXT_BORDER_WIDTH_Y + MUL_TEXT_BORDER_WIDTH_Y) - MUL_TEXT_BORDER_WIDTH_Y);

//		ul_corner.x = elem->corner.x + MUL_TEXT_BORDER_WIDTH_X;
//		ul_corner.y = ul_corner.y + MUL_TEXT_BORDER_WIDTH_Y;
//		lr_corner.x = elem->corner.x + elem->width - MUL_TEXT_BORDER_WIDTH_X;
//		lr_corner.y = lr_corner.y + MUL_TEXT_BORDER_WIDTH_Y;
		renderer->ops->set_fillstyle(renderer, FILLSTYLE_SOLID);
		renderer->ops->fill_rect(renderer,
				&ul_corner,
				&lr_corner,
				&task->inner_color);
		renderer->ops->set_linewidth(renderer, border_width);
		renderer->ops->set_linestyle(renderer, LINESTYLE_SOLID);
		renderer->ops->set_linejoin(renderer, LINEJOIN_MITER);
		renderer->ops->draw_rect(renderer,
				&ul_corner,
				&lr_corner,
				&task->border_color);

		//fonts
		p.x = elem->corner.x + elem->width / 2.0;;
		p.y = ul_corner.y + font_ascent(task->font, MUL_FONT_HEIGHT);
		renderer->ops->set_font(renderer,
				task->font,  MUL_FONT_HEIGHT);
		renderer->ops->draw_string(renderer,
				task->name, &p, ALIGN_CENTER,
				&color_black);

		ul_corner.y = ul_corner.y + MUL_FONT_HEIGHT;
		lr_corner.y = lr_corner.y + MUL_FONT_HEIGHT;

		tasks = g_slist_next(tasks);
		i++;
	}
	while(tasks);
}

//returns distance from point ;-)
static real mul_distance_from(Mul *mul, Point *point)
{
	Element *elem = &mul->element;
	Rectangle rect;

	rect.left = elem->corner.x - border_width/2;
	rect.right = elem->corner.x + elem->width + border_width/2;
	rect.top = elem->corner.y - border_width/2;
	rect.bottom = elem->corner.y + elem->height + border_width/2;

	return distance_rectangle_point(&rect, point);
}

//called when a mul is selected
static void mul_select(Mul *mul, Point *clicked_point,
		Renderer *interactive_renderer)
{
	element_update_handles(&mul->element);
}

//duplicates a mul
static Object* mul_copy(Mul *mul)
{
  int i;
  Mul *newmul;
  Element *elem, *newelem;
  Object *newobj;
  GSList *tasks, *newtasks;
  
  elem = &mul->element;
  
  newmul = g_malloc0(sizeof(Mul));
  newelem = &newmul->element;
  newobj = &newelem->object;

  element_copy(elem, newelem);

  for (i=0;i<8;i++) {
    newobj->connections[i] = &newmul->connections[i];
    newmul->connections[i].object = newobj;
    newmul->connections[i].connected = NULL;
    newmul->connections[i].pos = mul->connections[i].pos;
    newmul->connections[i].last_pos = mul->connections[i].last_pos;
  }

	tasks = mul->tasks;
	do
	{
		newtasks = g_slist_append(tasks, task_copy(tasks->data));
		tasks = g_slist_next(tasks);
	}
	while(tasks);
	newmul->tasks = newtasks;

	newmul->nulltask = NULL;

	newmul->border_color = mul->border_color;
	newmul->inner_color = mul->inner_color;

	newmul->properties_dialog = NULL;

	return &newmul->element.object;
}

//called when a mul is moved
static void mul_move(Mul *mul, Point *to)
{
	mul->element.corner = *to;
	
	mul_update_data(mul);
}

static void mul_move_handle(Mul *mul, Handle *handle,
		Point *to, HandleMoveReason reason, ModifierKeys modifiers)
{
	assert(mul!=NULL);
	assert(handle!=NULL);
	assert(to!=NULL);

	element_move_handle(&mul->element, handle->id, to, reason);

	mul_update_data(mul);
}

void mul_update_data(Mul *mul)
{
	int i;
	GSList *tasks;
	Task *task;

	Element *elem = &mul->element;
	Object *obj = &elem->object;
	ElementBBExtras *extra = &elem->extra_spacing;

	tasks = mul->tasks;

//	for(i = 0; i < 2; i++)
//	mul->task[i]->name_width = font_string_width(mul->task[i]->name, mul->font, MUL_FONT_HEIGHT);
	i = 0;

	//FIXME: this avoids resizing of the element
	//this possibility could be discussed, but I think this is better
	elem->width = 0;
	do
	{
		task = tasks->data;
		task->name_width = font_string_width(task->name, task->font, MUL_FONT_HEIGHT);
		elem->width = MAX(elem->width, task->name_width);
		tasks = g_slist_next(tasks);
		i++;
	}
	while(tasks);

	elem->width = elem->width + 2 * MUL_TEXT_BORDER_WIDTH_X;
	elem->height = i * MUL_FONT_HEIGHT + 2 * i * MUL_TEXT_BORDER_WIDTH_Y;

  /* Update connections: */
  mul->connections[0].pos = elem->corner;
  mul->connections[1].pos.x = elem->corner.x + elem->width / 2.0;
  mul->connections[1].pos.y = elem->corner.y;
  mul->connections[2].pos.x = elem->corner.x + elem->width;
  mul->connections[2].pos.y = elem->corner.y;
  mul->connections[3].pos.x = elem->corner.x;
  mul->connections[3].pos.y = elem->corner.y + elem->height / 2.0;
  mul->connections[4].pos.x = elem->corner.x + elem->width;
  mul->connections[4].pos.y = elem->corner.y + elem->height / 2.0;
  mul->connections[5].pos.x = elem->corner.x;
  mul->connections[5].pos.y = elem->corner.y + elem->height;
  mul->connections[6].pos.x = elem->corner.x + elem->width / 2.0;
  mul->connections[6].pos.y = elem->corner.y + elem->height;
  mul->connections[7].pos.x = elem->corner.x + elem->width;
  mul->connections[7].pos.y = elem->corner.y + elem->height;

  extra->border_trans = border_width/2.0;
  element_update_boundingbox(elem);
  
  obj->position = elem->corner;
  
  element_update_handles(elem);
}

//mul_types_ops
//creates a mul
static Object *mul_create(Point *startpoint, void *user_data,
		Handle **handle1,
		Handle **handle2)
{
	Mul *mul;
	Element *elem;
	Object *obj;
	int i;

	mul = g_malloc0(sizeof(Mul));
	elem = &mul->element;
	obj = &elem->object;

	obj->type = &mul_type;

	obj->ops = &mul_ops;

	elem->corner = *startpoint;
	elem->width = MUL_DEFAULT_WIDTH;
	elem->height = MUL_DEFAULT_HEIGHT;

//	mul->border_width =  attributes_get_default_linewidth();
	mul->border_color = attributes_get_foreground();
	mul->inner_color = attributes_get_background();

	element_init(elem, 8, 8);

	for(i = 0; i < 8; i++)
	{
		obj->connections[i] = &mul->connections[i];
		mul->connections[i].object = obj;
		mul->connections[i].connected = NULL;
	}

	mul->properties_dialog = NULL;

	mul->tasks = g_slist_append(mul->tasks, task_create(user_data));
	mul->tasks = g_slist_append(mul->tasks, task_create(user_data));
	mul->nulltask = NULL;

	mul_update_data(mul);

	for(i = 0; i < 8; i++)
	{
		obj->handles[i]->type = HANDLE_NON_MOVABLE;
	}

	*handle1 = NULL;
	*handle2 = obj->handles[0];

	return &mul->element.object;
}

//load a mul from a file
static Object* mul_load(ObjectNode obj_node, int version, const char *filename)
{
	Mul *mul;
	Element *elem;
	Object *obj;
	AttributeNode attr;
	int i;

	mul = g_new0(Mul, 1);
	elem = &mul->element;
	obj = &elem->object;

	obj->type = &mul_type;
	obj->ops = &mul_ops;

	element_load(elem, obj_node);

	//load number of tasks
	mul->nb_tasks = 2;
	attr = object_find_attribute(obj_node, "nb_tasks");
	if(attr != NULL)
		mul->nb_tasks = data_int(attribute_first_data(attr));

	for(i = 0; i < mul->nb_tasks; i++)
		mul->tasks = g_slist_append(mul->tasks, task_load(obj_node, version, filename, i));
//	mul->task[0] = task_load(obj_node, version, filename, 0);
//	mul->task[1] = task_load(obj_node, version, filename, 1);

	//load outer colours
	mul->border_color = color_black;
	attr = object_find_attribute(obj_node, "border_color");
	if(attr != NULL)
		data_color(attribute_first_data(attr), &mul->border_color);
	mul->inner_color = color_white;
	attr = object_find_attribute(obj_node, "inner_color");
	if (attr != NULL)
		data_color(attribute_first_data(attr), &mul->inner_color);

	mul->nulltask = NULL;

	element_init(elem, 8, 8);

	//load connections state
  for (i=0;i<8;i++) {
    obj->connections[i] = &mul->connections[i];
    mul->connections[i].object = obj;
    mul->connections[i].connected = NULL;
  }

/*  mul->name_width =
    font_string_width(mul->name, mul->font, MUL_FONT_HEIGHT);
*/
  mul->properties_dialog = NULL;

  mul_update_data(mul);

  for (i=0;i<8;i++) {
    obj->handles[i]->type = HANDLE_NON_MOVABLE;
  }

	return &mul->element.object;
}

//save a mul to a file
static void mul_save(Mul *mul, ObjectNode obj_node, const char *filename)
{
	GSList *tasks;

	mul->nb_tasks = 0;
	element_save(&mul->element, obj_node);

	tasks = mul->tasks;
	do
	{
		task_save(tasks->data, obj_node, filename, mul->nb_tasks);
		mul->nb_tasks++;
		tasks = g_slist_next(tasks);
	}
	while(tasks);
//	task_save(mul->task[0], obj_node, filename, 0);
//	task_save(mul->task[1], obj_node, filename, 1);

	data_add_int(new_attribute(obj_node, "nb_tasks"), mul->nb_tasks);
	data_add_color(new_attribute(obj_node, "border_color"), &mul->border_color);
	data_add_color(new_attribute(obj_node, "inner_color"), &mul->inner_color);
//	printf("WARNING: %s: line %d: not fully implemented yet!\n", __FILE__, __LINE__);
}
