/* Dia -- an diagram creation/manipulation program
 * Copyright (C) 1998 Alexander Larsson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <gtk/gtk.h>
#include <math.h>
//#include <string.h>

#include "intl.h"
#include "object.h"
#include "element.h"
#include "connectionpoint.h"
#include "render.h"
#include "attributes.h"
#include "widgets.h"
#include "properties.h"

#include "pixmaps/stask.xpm"

#include "cast.h"
#include "task.h"
#include "stask.h"
#include "stask_callbacks.h"



ObjectType stask_type;
ObjectType *_stask_type = (ObjectType *) &stask_type;


/* task functions */
/* task_ops */
static void stask_destroy(Stask *task);
static void stask_draw(Stask *stask, Renderer *renderer);
static real stask_distance_from(Stask *stask, Point *point);
static void stask_select(Stask *stask, Point *clicked_point, Renderer *interactive_renderer);
static Object *stask_copy(Stask *stask);
static void stask_move(Stask *stask, Point *to);
static void stask_move_handle(Stask *stask, Handle *handle, Point *to, HandleMoveReason reason, ModifierKeys modifiers);

//task_type_ops
static Object *stask_create(Point *startpoint, void *user_data, Handle **handle1, Handle **handle2);
static void stask_save(Stask *stask, ObjectNode obj_node, const char *filename);
static Object *stask_load(ObjectNode obj_node, int version, const char *filename);


//register task functions
//task_ops
static ObjectOps stask_ops = {
	(DestroyFunc)		stask_destroy,
	(DrawFunc)		stask_draw,
	(DistanceFunc)		stask_distance_from,
	(SelectFunc)		stask_select,
	(CopyFunc)		stask_copy,
	(MoveFunc)		stask_move,
	(MoveHandleFunc)	stask_move_handle,
	(GetPropertiesFunc)	stask_get_properties,
	(ApplyPropertiesFunc)	stask_apply_props,
	(ObjectMenuFunc)	NULL
//	(DescribePropsFunc)	stask_describe_props,
//	(GetPropsFunc)		stask_get_props,
//	(SetPropsFunc)		stask_set_props
};


//task_type_ops
static ObjectTypeOps stask_type_ops =
{
	(CreateFunc)	stask_create,
	(LoadFunc)	stask_load,
	(SaveFunc)	stask_save
};


//register task (cf cast.c)
ObjectType stask_type =
{
	"Cast - Task",		/* name */
	0,			/* version */
	(char **) stask_xpm,	/* pixmap */

	&stask_type_ops		/* ops */
};

//register task functions
//task_ops
//removes a task from memory
static void stask_destroy(Stask *stask)
{
	element_destroy(&stask->element);

	task_destroy(stask->task);

	if(stask->properties_dialog != NULL)
	{
		gtk_widget_destroy(stask->properties_dialog->dialog);
		g_free(stask->properties_dialog);
	}
}

//called to draw a task
static void stask_draw(Stask *stask, Renderer *renderer)
{
  Point ul_corner, lr_corner;
  Point p;
  Element *elem;

  assert(stask != NULL);
  assert(renderer != NULL);

  elem = &stask->element;

  ul_corner.x = elem->corner.x;
  ul_corner.y = elem->corner.y;
  lr_corner.x = elem->corner.x + elem->width;
  lr_corner.y = elem->corner.y + elem->height;
  
  renderer->ops->set_fillstyle(renderer, FILLSTYLE_SOLID);
  
  renderer->ops->fill_rect(renderer, 
			   &ul_corner,
			   &lr_corner, 
			   &stask->task->inner_color);

  renderer->ops->set_linewidth(renderer, border_width);
  renderer->ops->set_linestyle(renderer, LINESTYLE_SOLID);
  renderer->ops->set_linejoin(renderer, LINEJOIN_MITER);

  renderer->ops->draw_rect(renderer, 
			   &ul_corner,
			   &lr_corner, 
			   &stask->task->border_color);

  p.x = elem->corner.x + elem->width / 2.0;
  p.y = elem->corner.y + (elem->height - STASK_FONT_HEIGHT)/2.0 + font_ascent(stask->task->font, STASK_FONT_HEIGHT);
  renderer->ops->set_font(renderer, 
			  stask->task->font, STASK_FONT_HEIGHT);
  renderer->ops->draw_string(renderer, 
			     stask->task->name, 
			     &p, ALIGN_CENTER, 
			     &color_black);
}

//returns distance from point ;-)
static real stask_distance_from(Stask *stask, Point *point)
{
	Element *elem = &stask->element;
	Rectangle rect;

	rect.left = elem->corner.x - border_width/2;
	rect.right = elem->corner.x + elem->width + border_width/2;
	rect.top = elem->corner.y - border_width/2;
	rect.bottom = elem->corner.y + elem->height + border_width/2;

	return distance_rectangle_point(&rect, point);
}

//called when a task is selected
static void stask_select(Stask *stask, Point *clicked_point,
		Renderer *interactive_renderer)
{
	element_update_handles(&stask->element);
}

//duplicates a task
static Object * stask_copy(Stask *stask)
{
  int i;
  Stask *newstask;
  Element *elem, *newelem;
  Object *newobj;
  
  elem = &stask->element;
  
  newstask = g_malloc0(sizeof(Stask));
  newelem = &newstask->element;
  newobj = &newelem->object;

  element_copy(elem, newelem);

  for (i=0;i<8;i++) {
    newobj->connections[i] = &newstask->connections[i];
    newstask->connections[i].object = newobj;
    newstask->connections[i].connected = NULL;
    newstask->connections[i].pos = stask->connections[i].pos;
    newstask->connections[i].last_pos = stask->connections[i].last_pos;
  }

  newstask->task = task_copy(stask->task);

  newstask->properties_dialog = NULL;

  return &newstask->element.object;
}

//called when a task is moved
static void stask_move(Stask *stask, Point *to)
{
	stask->element.corner = *to;
	
	stask_update_data(stask);
}

static void stask_move_handle(Stask *stask, Handle *handle,
		Point *to, HandleMoveReason reason, ModifierKeys modifiers)
{
	assert(stask!=NULL);
	assert(handle!=NULL);
	assert(to!=NULL);

	element_move_handle(&stask->element, handle->id, to, reason);

	stask_update_data(stask);
}

//task_types_ops
//creates a task
static Object *stask_create(Point *startpoint, void *user_data,
		Handle **handle1,
		Handle **handle2)
{
  Stask *stask;
  Element *elem;
  Object *obj;
  int i;

  stask = g_malloc0(sizeof(Stask));
  elem = &stask->element;
  obj = &elem->object;
  
  obj->type = &stask_type;

  obj->ops = &stask_ops;

  elem->corner = *startpoint;
  elem->width = STASK_DEFAULT_WIDTH;
  elem->height = STASK_DEFAULT_HEIGHT;

  element_init(elem, 8, 8);

  for (i=0;i<8;i++) {
    obj->connections[i] = &stask->connections[i];
    stask->connections[i].object = obj;
    stask->connections[i].connected = NULL;
  }

  stask->properties_dialog = NULL;

  stask->task = task_create(user_data);

  stask_update_data(stask);

  for (i=0;i<8;i++) {
    obj->handles[i]->type = HANDLE_NON_MOVABLE;
  }

  *handle1 = NULL;
  *handle2 = obj->handles[0];  
  return &stask->element.object;
}

//load a task from a file
static Object* stask_load(ObjectNode obj_node, int version, const char *filename)
{
	Stask *stask;
	Element *elem;
	Object *obj;
	int i;

	stask = g_new0(Stask, 1);
	elem = &stask->element;
	obj = &elem->object;

	obj->type = &stask_type;
	obj->ops = &stask_ops;

	element_load(elem, obj_node);

	stask->task = task_load(obj_node, version, filename, 0);

	element_init(elem, 8, 8);

	//load connections state
  for (i=0;i<8;i++) {
    obj->connections[i] = &stask->connections[i];
    stask->connections[i].object = obj;
    stask->connections[i].connected = NULL;
  }

  stask->task->name_width =
    font_string_width(stask->task->name, stask->task->font, STASK_FONT_HEIGHT);

  stask->properties_dialog = NULL;

  for (i=0;i<8;i++) {
    obj->handles[i]->type = HANDLE_NON_MOVABLE;
  }

	return &stask->element.object;
}

//save a task to a file
static void stask_save(Stask *stask, ObjectNode obj_node, const char *filename)
{
	element_save(&stask->element, obj_node);

	task_save(stask->task, obj_node, filename, 0);
}

//common functions
void stask_update_data(Stask *stask)
{
  Element *elem = &stask->element;
  Object *obj = &elem->object;
  ElementBBExtras *extra = &elem->extra_spacing;

  stask->task->name_width = font_string_width(stask->task->name, stask->task->font, STASK_FONT_HEIGHT);

  elem->width = stask->task->name_width + 2 * STASK_TEXT_BORDER_WIDTH_X;
  elem->height = STASK_FONT_HEIGHT + 2 * STASK_TEXT_BORDER_WIDTH_Y;

  /* Update connections: */
  stask->connections[0].pos = elem->corner;
  stask->connections[1].pos.x = elem->corner.x + elem->width / 2.0;
  stask->connections[1].pos.y = elem->corner.y;
  stask->connections[2].pos.x = elem->corner.x + elem->width;
  stask->connections[2].pos.y = elem->corner.y;
  stask->connections[3].pos.x = elem->corner.x;
  stask->connections[3].pos.y = elem->corner.y + elem->height / 2.0;
  stask->connections[4].pos.x = elem->corner.x + elem->width;
  stask->connections[4].pos.y = elem->corner.y + elem->height / 2.0;
  stask->connections[5].pos.x = elem->corner.x;
  stask->connections[5].pos.y = elem->corner.y + elem->height;
  stask->connections[6].pos.x = elem->corner.x + elem->width / 2.0;
  stask->connections[6].pos.y = elem->corner.y + elem->height;
  stask->connections[7].pos.x = elem->corner.x + elem->width;
  stask->connections[7].pos.y = elem->corner.y + elem->height;

  extra->border_trans = border_width/2.0;
  element_update_boundingbox(elem);
  
  obj->position = elem->corner;
  
  element_update_handles(elem);
}
