/* Dia -- an diagram creation/manipulation program
 * Copyright (C) 1998 Alexander Larsson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>

#include "intl.h"
#include "object.h"
#include "element.h"
#include "connectionpoint.h"
#include "render.h"
#include "attributes.h"
#include "widgets.h"
#include "properties.h"

#include "pixmaps/sum.xpm"

#include "cast.h"
#include "task.h"
#include "sum.h"
#include "sum_callbacks.h"



ObjectType sum_type;
ObjectType *_sum_type = (ObjectType *) &sum_type;


//sum functions
//sum_ops
static void sum_destroy(Sum *sum);
static void sum_draw(Sum *sum, Renderer *renderer);
static real sum_distance_from(Sum *sum, Point *point);
static void sum_select(Sum *sum, Point *clicked_point, Renderer *interactive_renderer);
static Object *sum_copy(Sum *sum);
static void sum_move(Sum *sum, Point *to);
static void sum_move_handle(Sum *sum, Handle *handle, Point *to, HandleMoveReason reason, ModifierKeys modifiers);

//sum_type_ops
static Object *sum_create(Point *startpoint, void *user_data, Handle **handle1, Handle **handle2);
static void sum_save(Sum *sum, ObjectNode obj_node, const char *filename);
static Object *sum_load(ObjectNode obj_node, int version, const char *filename);


//register sum functions
//sum_ops
static ObjectOps sum_ops = {
	(DestroyFunc)		sum_destroy,
	(DrawFunc)		sum_draw,
	(DistanceFunc)		sum_distance_from,
	(SelectFunc)		sum_select,
	(CopyFunc)		sum_copy,
	(MoveFunc)		sum_move,
	(MoveHandleFunc)	sum_move_handle,
	(GetPropertiesFunc)	sum_get_props,
	(ApplyPropertiesFunc)	sum_apply_props,
	(ObjectMenuFunc)	NULL
//	(GetPropsFunc)		sum_get_props,
//	(SetPropsFunc)		sum_set_props
};


//sum_type_ops
static ObjectTypeOps sum_type_ops =
{
	(CreateFunc)	sum_create,
	(LoadFunc)	sum_load,
	(SaveFunc)	sum_save
};


//register sum (cf cast.c)
ObjectType sum_type =
{
	"Cast - Sum",		/* name */
	0,			/* version */
	(char **) sum_xpm,	/* pixmap */

	&sum_type_ops		/* ops */
};

//register sum functions
//sum_ops
//removes a sum from memory
static void sum_destroy(Sum *sum)
{
	element_destroy(&sum->element);

	task_destroy(sum->task[0]);
	task_destroy(sum->task[1]);

	if(sum->properties_dialog != NULL)
	{
		gtk_widget_destroy(sum->properties_dialog->dialog);
		g_free(sum->properties_dialog);
	}
}

//called to draw a sum
static void sum_draw(Sum *sum, Renderer *renderer)
	//TODO: optimize the drawing
{
	Point ul_corner, lr_corner;
	Point p;
	Element *elem;

	assert(sum != NULL);
	assert(renderer != NULL);

	elem = &sum->element;

	//rectangles
	//outer
	ul_corner.x = elem->corner.x;
	ul_corner.y = elem->corner.y;
	lr_corner.x = elem->corner.x + elem->width;
	lr_corner.y = elem->corner.y + elem->height;
	renderer->ops->set_fillstyle(renderer, FILLSTYLE_SOLID);
	renderer->ops->fill_rect(renderer,
			&ul_corner,
			&lr_corner,
			&sum->inner_color);
	renderer->ops->set_linewidth(renderer, border_width);
	renderer->ops->set_linestyle(renderer, LINESTYLE_SOLID);
	renderer->ops->set_linejoin(renderer, LINEJOIN_MITER);
	renderer->ops->draw_rect(renderer,
			&ul_corner,
			&lr_corner,
			&sum->border_color);

	//task 0
	ul_corner.x = elem->corner.x + SUM_TEXT_BORDER_WIDTH_X;
	ul_corner.y = elem->corner.y + SUM_TEXT_BORDER_WIDTH_Y;
	lr_corner.x = elem->corner.x + elem->width - SUM_TEXT_BORDER_WIDTH_X;
	lr_corner.y = elem->corner.y + (elem->height / 2.0) - SUM_TEXT_BORDER_WIDTH_Y;
	renderer->ops->set_fillstyle(renderer, FILLSTYLE_SOLID);
	renderer->ops->fill_rect(renderer,
			&ul_corner,
			&lr_corner,
			&sum->task[0]->inner_color);
	renderer->ops->set_linewidth(renderer, border_width);
	renderer->ops->set_linestyle(renderer, LINESTYLE_SOLID);
	renderer->ops->set_linejoin(renderer, LINEJOIN_MITER);
	renderer->ops->draw_rect(renderer,
			&ul_corner,
			&lr_corner,
			&sum->task[0]->border_color);

	//task 1
	ul_corner.x = elem->corner.x + SUM_TEXT_BORDER_WIDTH_X;
	ul_corner.y = elem->corner.y + (elem->height / 2.0) + SUM_TEXT_BORDER_WIDTH_Y;
	lr_corner.x = elem->corner.x + elem->width - SUM_TEXT_BORDER_WIDTH_X;
	lr_corner.y = elem->corner.y + elem->height - SUM_TEXT_BORDER_WIDTH_Y;
	renderer->ops->set_fillstyle(renderer, FILLSTYLE_SOLID);
	renderer->ops->fill_rect(renderer,
			&ul_corner,
			&lr_corner,
			&sum->task[1]->inner_color);
	renderer->ops->set_linewidth(renderer, border_width);
	renderer->ops->set_linestyle(renderer, LINESTYLE_SOLID);
	renderer->ops->set_linejoin(renderer, LINEJOIN_MITER);
	renderer->ops->draw_rect(renderer,
			&ul_corner,
			&lr_corner,
			&sum->task[1]->border_color);

	//fonts
	//sum
	p.x = elem->corner.x + elem->width / 2.0;
	p.y = elem->corner.y + (elem->height - SUM_FONT_HEIGHT)/2.0 + font_ascent(sum->font, SUM_FONT_HEIGHT);
	renderer->ops->set_font(renderer,
			sum->font, SUM_FONT_HEIGHT);
	renderer->ops->draw_string(renderer,
			"SUM",
			&p, ALIGN_CENTER,
			&color_black);
	//task 0
	p.x = elem->corner.x + elem->width / 2.0;
	p.y = elem->corner.y + font_ascent(sum->font, SUM_FONT_HEIGHT) + SUM_TEXT_BORDER_WIDTH_Y;
	renderer->ops->set_font(renderer,
			sum->font, SUM_FONT_HEIGHT);
	renderer->ops->draw_string(renderer,
			sum->task[0]->name,
			&p, ALIGN_CENTER,
			&color_black);

	//task 1
	 p.x = elem->corner.x + elem->width / 2.0;
	 p.y = elem->corner.y + elem->height / 2.0 + font_ascent(sum->font, SUM_FONT_HEIGHT) + SUM_TEXT_BORDER_WIDTH_Y;
	 renderer->ops->set_font(renderer,
			 sum->font, SUM_FONT_HEIGHT);
	 renderer->ops->draw_string(renderer,
			 sum->task[1]->name,
			 &p, ALIGN_CENTER,
			 &color_black);
}

//returns distance from point ;-)
static real sum_distance_from(Sum *sum, Point *point)
{
	Element *elem = &sum->element;
	Rectangle rect;

	rect.left = elem->corner.x - border_width/2;
	rect.right = elem->corner.x + elem->width + border_width/2;
	rect.top = elem->corner.y - border_width/2;
	rect.bottom = elem->corner.y + elem->height + border_width/2;

	return distance_rectangle_point(&rect, point);
}

//called when a sum is selected
static void sum_select(Sum *sum, Point *clicked_point,
		Renderer *interactive_renderer)
{
	element_update_handles(&sum->element);
}

//duplicates a sum
static Object* sum_copy(Sum *sum)
{
  int i;
  Sum *newsum;
  Element *elem, *newelem;
  Object *newobj;
  
  elem = &sum->element;
  
  newsum = g_malloc0(sizeof(Sum));
  newelem = &newsum->element;
  newobj = &newelem->object;

  element_copy(elem, newelem);

  for (i=0;i<8;i++) {
    newobj->connections[i] = &newsum->connections[i];
    newsum->connections[i].object = newobj;
    newsum->connections[i].connected = NULL;
    newsum->connections[i].pos = sum->connections[i].pos;
    newsum->connections[i].last_pos = sum->connections[i].last_pos;
  }

	newsum->font = sum->font;

	newsum->task[0] = task_copy(sum->task[0]);
	newsum->task[1] = task_copy(sum->task[1]);

	newsum->determinist = sum->determinist;
	newsum->rif2 = sum->rif2;

	newsum->border_color = sum->border_color;
	newsum->inner_color = sum->inner_color;

	newsum->properties_dialog = NULL;

	return &newsum->element.object;
}

//called when a sum is moved
static void sum_move(Sum *sum, Point *to)
{
	sum->element.corner = *to;
	
	sum_update_data(sum);
}

static void sum_move_handle(Sum *sum, Handle *handle,
		Point *to, HandleMoveReason reason, ModifierKeys modifiers)
{
	assert(sum!=NULL);
	assert(handle!=NULL);
	assert(to!=NULL);

	element_move_handle(&sum->element, handle->id, to, reason);

	sum_update_data(sum);
}

//sum_types_ops
//creates a sum
static Object *sum_create(Point *startpoint, void *user_data, Handle **handle1, Handle **handle2)
	//TODO: tif1, tif2, tthen, telse
{
	Sum *sum;
	Element *elem;
	Object *obj;
	int i;

	sum = g_malloc0(sizeof(Sum));
	elem = &sum->element;
	obj = &elem->object;

	obj->type = &sum_type;

	obj->ops = &sum_ops;

	elem->corner = *startpoint;
	elem->width = SUM_DEFAULT_WIDTH;
	elem->height = SUM_DEFAULT_HEIGHT;

	sum->border_color = attributes_get_foreground();
	sum->inner_color = attributes_get_background();

	element_init(elem, 8, 8);

	for(i = 0; i < 8; i++)
	{
		obj->connections[i] = &sum->connections[i];
		sum->connections[i].object = obj;
		sum->connections[i].connected = NULL;
	}

	sum->font = font_getfont("Courier");

	sum->properties_dialog = NULL;

	sum->task[0] = task_create(user_data);
	sum->task[1] = task_create(user_data);

	sum->determinist = 0;
	sum->tif1 = sum->task[0];
	sum->op = OP_EQ;
	sum->tif2 = sum->task[1];
	sum->value = 0;
	sum->rif2 = 0;
	sum->tthen = sum->task[0];
	sum->telse = sum->task[1];

	sum_update_data(sum);

	for(i = 0; i < 8; i++)
	{
		obj->handles[i]->type = HANDLE_NON_MOVABLE;
	}

	*handle1 = NULL;
	*handle2 = obj->handles[0];

	return &sum->element.object;
}

//load a sum from a file
static Object* sum_load(ObjectNode obj_node, int version, const char *filename)
	//TODO: tif1, tif2, tthen, telse
{
	Sum *sum;
	Element *elem;
	Object *obj;
	AttributeNode attr;
	int i;

	sum = g_new0(Sum, 1);
	elem = &sum->element;
	obj = &elem->object;

	obj->type = &sum_type;
	obj->ops = &sum_ops;

	element_load(elem, obj_node);

	sum->task[0] = task_load(obj_node, version, filename, 0);
	sum->task[1] = task_load(obj_node, version, filename, 1);

	//determinist
	sum->determinist = 0;
	attr = object_find_attribute(obj_node, "determinist");
	if(attr != NULL)
		sum->determinist = data_boolean(attribute_first_data(attr));

	sum->tif1 = sum->task[0];
	attr = object_find_attribute(obj_node, "tif1");
	if(attr != NULL)
		sum->tif1 = sum->task[!data_boolean(attribute_first_data(attr))];
/*	{
		if(data_boolean(attribute_first_data(attr)))
			sum->tif1 = sum->task[0];
		else
			sum->tif1 = sum->task[1];
	}
*/
	sum->op = OP_EQ;
	attr = object_find_attribute(obj_node, "op");
	if(attr != NULL)
		sum->op = data_enum(attribute_first_data(attr));

	sum->tif2 = sum->task[0];
	attr = object_find_attribute(obj_node, "tif2");
	if(attr != NULL)
		sum->tif2 = sum->task[!data_boolean(attribute_first_data(attr))];

	sum->value = 0;
	attr = object_find_attribute(obj_node, "value");
	if(attr != NULL)
		sum->value = data_boolean(attribute_first_data(attr));

	sum->rif2 = 0.0;
	attr = object_find_attribute(obj_node, "rif2");
	if(attr != NULL)
		sum->rif2 = data_real(attribute_first_data(attr));

	sum->tthen = sum->task[0];
	attr = object_find_attribute(obj_node, "tthen");
	if(attr != NULL)
		sum->tthen = sum->task[!data_boolean(attribute_first_data(attr))];

	sum->telse = sum->task[0];
	attr = object_find_attribute(obj_node, "telse");
	if(attr != NULL)
		sum->telse = sum->task[!data_boolean(attribute_first_data(attr))];

	//load outer colours
	sum->border_color = color_black;
	attr = object_find_attribute(obj_node, "border_color");
	if(attr != NULL)
		data_color(attribute_first_data(attr), &sum->border_color);
	sum->inner_color = color_white;
	attr = object_find_attribute(obj_node, "inner_color");
	if (attr != NULL)
		data_color(attribute_first_data(attr), &sum->inner_color);

	sum->font = font_getfont(_("Courier"));

	element_init(elem, 8, 8);

	//load connections state
  for (i=0;i<8;i++) {
    obj->connections[i] = &sum->connections[i];
    sum->connections[i].object = obj;
    sum->connections[i].connected = NULL;
  }

/*  sum->name_width =
    font_string_width(sum->name, sum->font, FONT_HEIGHT);
*/
  sum->properties_dialog = NULL;

  sum_update_data(sum);

  for (i=0;i<8;i++) {
    obj->handles[i]->type = HANDLE_NON_MOVABLE;
  }

	return &sum->element.object;
}

//save a sum to a file
static void sum_save(Sum *sum, ObjectNode obj_node, const char *filename)
{
	element_save(&sum->element, obj_node);

	//the tasks
	task_save(sum->task[0], obj_node, filename, 0);
	task_save(sum->task[1], obj_node, filename, 1);

	//determinist
	data_add_boolean(new_attribute(obj_node, "determinist"), sum->determinist);

//	printf("sum_save: %d\n", sum->tif1 == sum->task[0]);
	data_add_boolean(new_attribute(obj_node, "tif1"), sum->tif1 == sum->task[0]);
/*	if(sum->tif1 == sum->task[0])
		data_add_string(new_attribute(obj_node, "tif1"), sum->task[0]->name);
	else if(sum->tif1 == sum->task[1])
		data_add_string(new_attribute(obj_node, "tif1"), sum->task[1]->name);
*/
	data_add_enum(new_attribute(obj_node, "op"), sum->op);

	data_add_boolean(new_attribute(obj_node, "tif2"), sum->tif2 == sum->task[0]);
	data_add_boolean(new_attribute(obj_node, "value"), sum->value);
/*
	if(sum->tif2 == sum->task[0])
		data_add_string(new_attribute(obj_node, "tif2"), sum->task[0]->name);
	else if(sum->tif2 == sum->task[1])
		data_add_string(new_attribute(obj_node, "tif2"), sum->task[1]->name);
*/
	data_add_real(new_attribute(obj_node, "rif2"), sum->rif2);

	data_add_boolean(new_attribute(obj_node, "tthen"), sum->tthen == sum->task[0]);
	data_add_boolean(new_attribute(obj_node, "telse"), sum->telse == sum->task[0]);
/*
	if(sum->tthen == sum->task[0])
		data_add_string(new_attribute(obj_node, "tthen"), sum->task[0]->name);
	else if(sum->tthen == sum->task[1])
		data_add_string(new_attribute(obj_node, "tthen"), sum->task[1]->name);

	if(sum->telse == sum->task[0])
		data_add_string(new_attribute(obj_node, "telse"), sum->task[0]->name);
	else if(sum->telse == sum->task[1])
		data_add_string(new_attribute(obj_node, "telse"), sum->task[1]->name);
*/
	data_add_color(new_attribute(obj_node, "border_color"), &sum->border_color);
	data_add_color(new_attribute(obj_node, "inner_color"), &sum->inner_color);
	data_add_font(new_attribute(obj_node, "font"), sum->font);
}

//common fonctions
void sum_update_data(Sum *sum)
{
	int i;
	Element *elem = &sum->element;
	Object *obj = &elem->object;
	ElementBBExtras *extra = &elem->extra_spacing;

	for(i = 0; i < 2; i++)
		sum->task[i]->name_width = font_string_width(sum->task[i]->name, sum->font, SUM_FONT_HEIGHT);

	elem->width = MAX(sum->task[0]->name_width, sum->task[1]->name_width) + 2 * SUM_TEXT_BORDER_WIDTH_X;
	elem->height = 2 * SUM_FONT_HEIGHT + 4 * SUM_TEXT_BORDER_WIDTH_Y;

  /* Update connections: */
  sum->connections[0].pos = elem->corner;
  sum->connections[1].pos.x = elem->corner.x + elem->width / 2.0;
  sum->connections[1].pos.y = elem->corner.y;
  sum->connections[2].pos.x = elem->corner.x + elem->width;
  sum->connections[2].pos.y = elem->corner.y;
  sum->connections[3].pos.x = elem->corner.x;
  sum->connections[3].pos.y = elem->corner.y + elem->height / 2.0;
  sum->connections[4].pos.x = elem->corner.x + elem->width;
  sum->connections[4].pos.y = elem->corner.y + elem->height / 2.0;
  sum->connections[5].pos.x = elem->corner.x;
  sum->connections[5].pos.y = elem->corner.y + elem->height;
  sum->connections[6].pos.x = elem->corner.x + elem->width / 2.0;
  sum->connections[6].pos.y = elem->corner.y + elem->height;
  sum->connections[7].pos.x = elem->corner.x + elem->width;
  sum->connections[7].pos.y = elem->corner.y + elem->height;

  extra->border_trans = border_width/2.0;
  element_update_boundingbox(elem);
  
  obj->position = elem->corner;
  
  element_update_handles(elem);

  //FIXME: name_width
}
