/* Dia -- an diagram creation/manipulation program
 * Copyright (C) 1998 Alexander Larsson
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>

#include "intl.h"
#include "object.h"
#include "element.h"
#include "connectionpoint.h"
#include "render.h"
#include "attributes.h"
#include "widgets.h"
#include "properties.h"

#include "cast.h"
#include "task.h"

//task functions
//utility functions
const char default_name[] = "Task %d";
static short task_counter = 1;
static char* task_new_name(void);


//task functions
//utility functions
static char* task_new_name(void)
{
	char str[sizeof(default_name) + 2];
	sprintf(str, default_name, task_counter++);
	return g_strdup(str);
}

//task_ops
//removes a task from memory
void task_destroy(Task *task)
{
	g_free(task->name);
	g_free(task->servername);
	g_free(task->corbaobjectname);
	g_free(task->corbamethodname);

	if(task->modules_window != NULL)
	{
		gtk_widget_destroy(task->modules_window->window);
		g_free(task->modules_window);
	}
}

//duplicates a task
Task* task_copy(Task *task)
{
	Task *newtask;

	newtask = g_malloc0(sizeof(Task));

//	newtask->name = task_new_name();
	newtask->name = g_strdup(task->name);
	newtask->font = task->font;
	newtask->name_width = task->name_width;

	newtask->type = task->type;

	newtask->properties_dialog = task->properties_dialog;
	newtask->modules_window = NULL;

	if(newtask->servername) newtask->servername = g_strdup(task->servername);
	newtask->usescorba = task->usescorba;
	if(newtask->corbaobjectname) newtask->corbaobjectname = g_strdup(task->corbaobjectname);
	if(newtask->corbamethodname) newtask->corbamethodname = g_strdup(task->corbamethodname);
	newtask->usespaco = task->usespaco;

	newtask->border_color = task->border_color;
	newtask->inner_color = task->inner_color;
	
	return newtask;
}


//task_types_ops
//creates a task
//Task* task_create(Point *startpoint, void *user_data, Handle **handle1, Handle **handle2)
Task* task_create(void *user_data)
{
	Task *task;

	task = g_new0(Task, 1);

	task->name = task_new_name();
	task->font = font_getfont("Courier");
	task->name_width = font_string_width(task->name, task->font, TASK_FONT_HEIGHT);
	task->type = GPOINTER_TO_INT(user_data);
	task->servername = NULL;
	task->modules_window = NULL;
	task->usescorba = FALSE;
	task->corbaobjectname = NULL;
	task->corbamethodname = NULL;
	task->usespaco = FALSE;

	task->border_color = attributes_get_foreground();
	task->inner_color = attributes_get_background();

	return task;
}

//load a task from a file
Task* task_load(ObjectNode obj_node, int version, const char *filename, short i)
{
	Task *task;
	AttributeNode attr;

	task = g_new0(Task, 1);

	task->modules_window = NULL;

	//load name
	task->name = NULL;
	attr = object_find_attribute(obj_node, cast_get_name("name", i));
	if (attr != NULL)
		task->name = data_string(attribute_first_data(attr));

	//load font
	//FIXME: do we need to load a font? setting another is not implemented yet
	task->font = font_getfont(_("Courier"));

	//load type
	task->type = TASK_COMPOSITE;
	attr = object_find_attribute(obj_node, cast_get_name("type", i));
	if (attr != NULL)
		task->type = (TaskType)data_enum(attribute_first_data(attr));

	//load the servername field
	task->servername = NULL;
	attr = object_find_attribute(obj_node, cast_get_name("servername", i));
	if(attr != NULL)
		task->servername = data_string(attribute_first_data(attr));
	
	//do we use corba?
	task->usescorba = TASK_WITHOUT_CORBA;
	attr = object_find_attribute(obj_node, cast_get_name("usescorba", i));
	if(attr != NULL)
		task->usescorba = data_boolean(attribute_first_data(attr));

	//load corba properties
	task->corbaobjectname = NULL;
	attr = object_find_attribute(obj_node, cast_get_name("corbaobjectname", i));
	if(attr != NULL)
		task->corbaobjectname = data_string(attribute_first_data(attr));
	task->corbamethodname = NULL;
	attr = object_find_attribute(obj_node, cast_get_name("corbamethodname", i));
	if(attr != NULL)
		task->corbamethodname = data_string(attribute_first_data(attr));
	task->usespaco = FALSE;
	attr = object_find_attribute(obj_node, cast_get_name("usespaco", i));
	if(attr != NULL)
		task->usespaco = data_boolean(attribute_first_data(attr));

	//load colours
	task->border_color = color_black;
	attr = object_find_attribute(obj_node, cast_get_name("border_color", i));
	if (attr != NULL)
		data_color(attribute_first_data(attr), &task->border_color);
	task->inner_color = color_white;
	attr = object_find_attribute(obj_node, cast_get_name("inner_color", i));
	if (attr != NULL)
		data_color(attribute_first_data(attr), &task->inner_color);

	return task;
}

//save a task to a file
void task_save(Task *task, ObjectNode obj_node, const char *filename, short i)
{
	data_add_string(new_attribute(obj_node, cast_get_name("name", i)), task->name);
	data_add_enum(new_attribute(obj_node, cast_get_name("type", i)), task->type);

	data_add_string(new_attribute(obj_node, cast_get_name("servername", i)), task->servername);

	//CORBA properties
	data_add_boolean(new_attribute(obj_node, cast_get_name("usescorba", i)), task->usescorba);
	data_add_string(new_attribute(obj_node, cast_get_name("corbaobjectname", i)), task->corbaobjectname);
	data_add_string(new_attribute(obj_node, cast_get_name("corbamethodname", i)), task->corbamethodname);
	data_add_boolean(new_attribute(obj_node, cast_get_name("usespaco", i)), task->usespaco);

	//colors
	data_add_color(new_attribute(obj_node, cast_get_name("border_color", i)), &task->border_color);
	data_add_color(new_attribute(obj_node, cast_get_name("inner_color", i)), &task->inner_color);
}
